import { doesExist, mustDefault, mustExist, NotImplementedError } from '@apextoaster/js-utils';
import {
  ChannelType,
  Client,
  Events,
  FormattingPatterns,
  GatewayIntentBits,
  Message as DiscordMessage,
  Partials,
  TextBasedChannel,
} from 'discord.js';

import { CommandContext } from '../command/index.js';
import { catchAndLog } from '../utils/async.js';
import { authorName, matchCommands } from '../utils/command.js';
import { Author, Bot, BotContext, Channel, Message } from './index.js';

export async function onMessage(context: CommandContext, client: Client<true>, discordMessage: DiscordMessage): Promise<void> {
  const { args, commands, logger } = context;
  const { author, content } = discordMessage;
  const name = authorName(author);

  if (name === mustExist(client.user).tag) {
    logger.debug('own message, ignoring');
    return;
  }

  const matchingCommands = matchCommands(content, commands, args.commandPrefix);
  const message = await messageFromDiscord(client, discordMessage);

  logger.debug({ message }, 'message received');

  for (const command of matchingCommands) {
    logger.debug({
      command: command.name,
    }, 'message matched command');

    if (doesExist(command.auth)) {
      const allowed = await command.auth(context, author);
      logger.debug({ allowed, author, command }, 'command auth check');

      if (allowed === false) {
        logger.warn({ author, command }, 'author attempted to use unauthorized command');
        continue;
      }
    }

    command.run(context, message).catch((err) => {
      logger.error(err, 'error executing command');
    });
  }
}

async function channelFromDiscord(_client: Client, channel: TextBasedChannel): Promise<Channel> {
  switch (channel.type) {
    case ChannelType.GuildAnnouncement:
    case ChannelType.GuildText:
      return {
        name: channel.name,
        server: channel.guild.name,
        topic: mustDefault(channel.topic, ''),
        type: 'text',
        async edit(options) {
          const edited = await channel.edit(options);
          return channelFromDiscord(_client, edited);
        }
      };
    case ChannelType.DM:
      return {
        name: channel.id,
        type: 'dm',
      };
    case ChannelType.PrivateThread:
    case ChannelType.PublicThread:
    default:
      throw new Error('invalid channel type');
  }
}

async function messageFromDiscord(client: Client, message: DiscordMessage): Promise<Message> {
  const channel = await getChannel(client, message.channelId);

  return {
    author: {
      discriminator: message.author.discriminator,
      username: message.author.username,
    },
    channel,
    content: message.content,
    async reply(response: string) {
      const reply = await message.reply(response);
      return messageFromDiscord(client, reply);
    },
  };
}

async function getChannel(client: Client, name: string): Promise<Channel> {
  const channel = await client.channels.fetch(name);
  if (doesExist(channel)) {
    // TODO: find a nice way of allowing more channel types
    if (channel.type === ChannelType.GuildText || channel.type === ChannelType.DM) {
      return channelFromDiscord(client, channel);
    } else {
      throw new Error('invalid channel type');
    }
  } else {
    throw new Error('channel does not exist');
  }
}

async function getUser(client: Client, ident: string): Promise<Author> {
  const match = FormattingPatterns.User.exec(ident);

  if (match) {
    const [_full, snowflake] = Array.from(match);
    const user = await client.users.fetch(snowflake);
    return {
      discriminator: user.discriminator,
      username: user.username,
    };
  } else {
    throw new Error('invalid user ident');
  }
}

/**
 * Connect to a Discord server using a bot token.
 *
 * Requires {@link ParsedArgs.discordToken} to be set in `args`.
 *
 * @public
 */
export async function discordConnect(botContext: BotContext, clientCtor: typeof Client = Client): Promise<Bot> {
  const { args, logger } = botContext;

  const client = new clientCtor({
    intents: [
      GatewayIntentBits.DirectMessages,
      GatewayIntentBits.DirectMessageReactions,
      GatewayIntentBits.Guilds,
      GatewayIntentBits.GuildMessages,
      GatewayIntentBits.GuildMessageReactions,
      GatewayIntentBits.MessageContent,
    ],
    partials: [
      Partials.Channel,
    ],
  });

  const bot: Bot = {
    async connect() {
      await client.login(args.discordToken);
      return true;
    },
    destroy() {
      client.destroy();
    },
    getChannel(name: string): Promise<Channel> {
      return getChannel(client, name);
    },
    getUser(name: string): Promise<Author> {
      return getUser(client, name);
    },
  };

  client.once(Events.ClientReady, (c) => {
    logger.info({
      user: c.user.tag,
    }, 'logged in and ready');
  });

  client.on(Events.MessageCreate, (message) => catchAndLog(onMessage({
    ...botContext,
    bot,
  }, client, message), logger, 'error handling message'));

  return bot;
}
