import { Logger } from 'noicejs';

import { ParsedArgs } from '../args.js';
import { Command } from '../command/index.js';
import { DataLayer } from '../data/index.js';
import { RconClient } from '../utils/rcon.js';
import { discordConnect } from './discord.js';
import { readlineConnect } from './readline.js';

/**
 * Message author.
 *
 * @public
 */
export interface Author {
  discriminator: string;
  username: string;
}

export type ChannelType = 'dm' | 'text';

export interface BaseChannel {
  name: string;
  type: ChannelType;
}

export interface TextChannel extends BaseChannel {
  server: string;
  topic: string;
  type: 'text';

  edit(options: {
    name?: string;
    topic?: string;
  }): Promise<Channel>;
}

export interface DirectChannel extends BaseChannel {
  type: 'dm';
}

export type Channel = DirectChannel | TextChannel;

/**
 * Text message with ability to reply.
 *
 * @public
 */
export interface Message {
  author: Author;
  channel: Channel;
  content: string;

  reply(message: string): Promise<Message>;
}

/**
 * Necessary methods for a bot.
 *
 * @public
 */
export interface Bot {
  connect(): Promise<boolean>;
  destroy(): void;

  getChannel(name: string): Promise<Channel | undefined>;
  getUser(name: string): Promise<Author | undefined>;
}

/**
 * Constructor for a Bot.
 *
 * @public
 */
export type BotCtor = (context: BotContext) => Promise<Bot>;

/**
 * Context passed to a {@link Bot} when it is created.
 *
 * @public
 */
export interface BotContext {
  args: ParsedArgs;
  commands: ReadonlyArray<Command>;
  data: DataLayer;
  logger: Logger;
  rcon: RconClient;
}

/**
 * Available bots.
 *
 * @private
 */
export const BOTS: {[key: string]: BotCtor} = {
  discord: discordConnect,
  readline: readlineConnect,
};
