import sqlite3 from 'sqlite3';

import { ParsedArgs } from '../args.js';
import { Author } from '../bot/index.js';
import { authorName } from '../utils/command.js';
import { DataLayer, DataTable, DataValue } from './index.js';

interface DataValueRow {
  value: string;
}

const { Database } = sqlite3;

export const SQLITE_SCHEMA = `
CREATE TABLE IF NOT EXISTS state (key TEXT PRIMARY KEY, value TEXT);
CREATE TABLE IF NOT EXISTS secret (key TEXT PRIMARY KEY, value TEXT);
`;

export class SqliteDataLayer implements DataLayer {
  protected db: sqlite3.Database;

  constructor(args: ParsedArgs, databaseCtor: typeof Database = Database) {
    this.db = new databaseCtor(args.sqliteDatabase);
  }

  public async connect(): Promise<void> {
    return new Promise((res, rej) => {
      this.db.exec(SQLITE_SCHEMA, (err) => {
        if (err) {
          rej(err);
        } else {
          res();
        }
      });
    });
  }

  public async key(scope: DataTable, author: Author, key: string) {
    return [
      scope,
      authorName(author),
      key,
    ].join(':');
  }

  public async get(scope: DataTable, author: Author, key: string): Promise<DataValue | undefined> {
    const compoundKey = await this.key(scope, author, key);
    // console.log('sqlite get', compoundKey);

    return new Promise((res, rej) => {
      this.db.get(`SELECT value FROM ${scope} WHERE key = ? LIMIT 1`, [
        compoundKey,
      ], (err: Error | undefined, row: DataValueRow | undefined) => {
        if (err) {
          rej(err);
        } else {
          if (row) {
            // console.log('sqlite get value', compoundKey, row);
            const value = JSON.parse(row.value) as DataValue;
            res(value);
          } else {
            res(undefined);
          }
        }
      });
    });
  }

  public async set(scope: DataTable, author: Author, key: string, value: DataValue): Promise<void> {
    const compoundKey = await this.key(scope, author, key);
    const jsonValue = JSON.stringify(value);
    // console.log('sqlite set', compoundKey, jsonValue);

    return new Promise((res, rej) => {
      this.db.run(`INSERT OR REPLACE INTO ${scope} (key, value) VALUES (?, ?)`, [
        compoundKey,
        jsonValue,
      ], (err) => {
        if (err) {
          rej(err);
        } else {
          // console.log('sqlite set done');
          res();
        }
      });
    });
  }
}
